import os
import sys

from dagster_graphql.test.utils import execute_dagster_graphql

from dagster import check
from dagster.core.scheduler import RunningSchedule
from dagster.seven import mock

from .utils import define_context

try:
    # Python 2 tempfile doesn't have tempfile.TemporaryDirectory
    import backports.tempfile as tempfile
except ImportError:
    import tempfile

GET_SCHEDULES_QUERY = '''
{
  scheduler {
        schedulerType
        schedules {
            scheduleId
        }
    }
}
'''


START_SCHEDULE_QUERY = '''
mutation StartSchedule($scheduleId: String!) {
  startRunSchedule(scheduleId: $scheduleId) {
    schedule {
      scheduleId
      pythonPath
      repositoryPath
    }
  }
}
'''

END_SCHEDULE_QUERY = '''
mutation EndSchedule($scheduleId: String!) {
  endRunSchedule(scheduleId: $scheduleId) {
    schedule {
      scheduleId
    }
  }
}
'''


def default_execution_params():
    return {
        "environmentConfigData": {"storage": {"filesystem": None}},
        "selector": {"name": "no_config_pipeline", "solidSubset": None},
        "mode": "default",
    }


def get_schedule_definition_from_context(context):
    repository = context.get_handle().build_repository_definition()
    schedule_def = repository.get_schedule("no_config_pipeline_hourly_schedule")

    return schedule_def


@mock.patch.dict(os.environ, {"DAGSTER_HOME": "~/dagster"})
def test_start_and_end_schedule():
    with tempfile.TemporaryDirectory() as schedule_dir:
        context = define_context(schedule_dir=schedule_dir)

        # Start schedule
        schedule_def = get_schedule_definition_from_context(context)
        scheduler = context.scheduler
        schedule = scheduler.start_schedule(schedule_def, sys.executable, "")

        check.inst_param(schedule, 'schedule', RunningSchedule)
        assert schedule.schedule_definition == schedule_def
        assert "/bin/python" in schedule.python_path

        assert "{}_{}.json".format(schedule_def.name, schedule.schedule_id) in os.listdir(
            schedule_dir
        )
        assert "{}_{}.sh".format(schedule_def.name, schedule.schedule_id) in os.listdir(
            schedule_dir
        )

        # End schedule
        scheduler.end_schedule(schedule_def)
        assert "{}_{}.json".format(schedule_def.name, schedule.schedule_id) not in os.listdir(
            schedule_dir
        )
        assert "{}_{}.sh".format(schedule_def.name, schedule.schedule_id) not in os.listdir(
            schedule_dir
        )


@mock.patch.dict(os.environ, {"DAGSTER_HOME": "~/dagster"})
def test_get_all_schedules():
    with tempfile.TemporaryDirectory() as schedule_dir:
        context = define_context(schedule_dir=schedule_dir)

        # Start schedule
        schedule_def = get_schedule_definition_from_context(context)
        scheduler = context.scheduler
        schedule = scheduler.start_schedule(schedule_def, sys.executable, "")

        # Query Scheduler + all Schedules
        scheduler_result = execute_dagster_graphql(
            context, GET_SCHEDULES_QUERY, variables={"pipelineName": "no_config_pipeline"}
        )

        assert scheduler_result.data
        assert scheduler_result.data['scheduler']
        assert scheduler_result.data['scheduler']['schedules']
        assert len(scheduler_result.data['scheduler']['schedules']) == 1

        assert (
            scheduler_result.data['scheduler']['schedules'][0]['scheduleId'] == schedule.schedule_id
        )



import json
import sys

from dagster_graphql import dauphin

from dagster import ScheduleDefinition, check
from dagster.core.scheduler import RunningSchedule, SystemCronScheduler


def get_scheduler(graphene_info):
    scheduler = graphene_info.context.scheduler

    if isinstance(scheduler, SystemCronScheduler):
        scheduler_type = SchedulerType.SystemCronScheduler
    else:
        raise Exception(
            'Unknown Scheduler type {typ}. Add this Scheduler type to the SchedulerType '
            'GraphQL Enum'.format(typ=type(scheduler))
        )

    schedules = [
        graphene_info.schema.type_named('RunningSchedule')(graphene_info, schedule=s)
        for s in scheduler.all_schedules()
    ]

    return graphene_info.schema.type_named('Scheduler')(
        scheduler_type=scheduler_type, schedules=schedules
    )


class DauphinScheduleDefinition(dauphin.ObjectType):
    class Meta:
        name = 'ScheduleDefinition'

    name = dauphin.NonNull(dauphin.String)
    cron_schedule = dauphin.NonNull(dauphin.String)
    execution_params_string = dauphin.NonNull(dauphin.String)

    def __init__(self, schedule_def):
        self._schedule_def = check.inst_param(schedule_def, 'schedule_def', ScheduleDefinition)

        super(DauphinScheduleDefinition, self).__init__(
            name=schedule_def.name,
            cron_schedule=schedule_def.cron_schedule,
            execution_params_string=json.dumps(schedule_def.execution_params),
        )


class DauphinRunningSchedule(dauphin.ObjectType):
    class Meta:
        name = 'RunningSchedule'

    schedule_id = dauphin.NonNull(dauphin.String)
    schedule_definition = dauphin.NonNull('ScheduleDefinition')
    python_path = dauphin.Field(dauphin.String)
    repository_path = dauphin.Field(dauphin.String)

    def __init__(self, graphene_info, schedule):
        self._schedule = check.inst_param(schedule, 'schedule', RunningSchedule)

        super(DauphinRunningSchedule, self).__init__(
            schedule_id=schedule.schedule_id,
            schedule_definition=graphene_info.schema.type_named('ScheduleDefinition')(
                schedule.schedule_definition
            ),
            python_path=schedule.python_path,
            repository_path=schedule.repository_path,
        )


class SchedulerType(dauphin.Enum):
    SystemCronScheduler = "SystemCronScheduler"


class DauphinScheduler(dauphin.ObjectType):
    class Meta:
        name = 'Scheduler'

    scheduler_type = dauphin.NonNull(dauphin.String)
    schedules = dauphin.non_null_list('RunningSchedule')


class DauphinRunningScheduleInput(dauphin.InputObjectType):
    class Meta:
        name = 'RunningScheduleInput'

    name = dauphin.NonNull(dauphin.String)
    cron_schedule = dauphin.NonNull(dauphin.String)
    execution_params = dauphin.NonNull('ExecutionParams')


class RunningScheduleResult(dauphin.ObjectType):
    class Meta:
        name = 'RunningScheduleResult'

    schedule = dauphin.NonNull('RunningSchedule')


class DauphinDeleteRunningScheduleResult(dauphin.ObjectType):
    class Meta:
        name = 'DeleteRunningScheduleResult'

    deleted_schedule = dauphin.NonNull('RunningSchedule')


class DauphinStartScheduleMutation(dauphin.Mutation):
    class Meta:
        name = 'StartScheduleMutation'

    class Arguments:
        schedule_id = dauphin.NonNull(dauphin.String)

    Output = dauphin.NonNull('RunningScheduleResult')

    def mutate(self, graphene_info, schedule_id):
        scheduler = graphene_info.context.scheduler

        python_path = sys.executable
        handle = graphene_info.context.get_handle()
        repository_path = handle.data.repository_yaml

        schedule = scheduler.start_schedule(schedule_id, python_path, repository_path)

        return graphene_info.schema.type_named('RunningScheduleResult')(
            schedule=graphene_info.schema.type_named('RunningSchedule')(schedule)
        )


class DauphinEndRunningScheduleMutation(dauphin.Mutation):
    class Meta:
        name = 'EndRunningScheduleMutation'

    class Arguments:
        schedule_id = dauphin.NonNull(dauphin.String)

    Output = dauphin.NonNull('RunningScheduleResult')

    def mutate(self, graphene_info, schedule_id):
        scheduler = graphene_info.context.scheduler

        schedule = scheduler.end_schedule(schedule_id)

        return graphene_info.schema.type_named('RunningScheduleResult')(
            schedule=graphene_info.schema.type_named('RunningSchedule')(schedule)
        )



import atexit

import gevent
import gevent.lock

from dagster.core.storage.event_log import EventLogSequence


class PipelineRunObservableSubscribe(object):
    def __init__(self, pipeline_run, after_cursor=None):
        self.event_log_sequence = EventLogSequence()
        self.pipeline_run = pipeline_run
        self.observer = None
        self.after_cursor = after_cursor or -1
        self.lock = gevent.lock.Semaphore()
        self.flush_scheduled = False
        self.flush_after = 0.75
        atexit.register(self._cleanup)

    def __call__(self, observer):
        self.observer = observer
        events = self.pipeline_run.logs_after(self.after_cursor)
        if events:
            self.observer.on_next(events)
        self.pipeline_run.subscribe(self)

    def handle_new_event(self, new_event):
        with self.lock:
            self.event_log_sequence = self.event_log_sequence.append(new_event)

            if self.flush_after is None:
                self.observer.on_next(self.event_log_sequence)
                self.event_log_sequence = EventLogSequence()
                return

            if not self.flush_scheduled:
                self.flush_scheduled = True
                gevent.spawn(self._flush_logs_after_delay)

    def _flush_logs_after_delay(self):
        gevent.sleep(self.flush_after)
        with self.lock:
            self.observer.on_next(self.event_log_sequence)
            self.event_log_sequence = EventLogSequence()
            self.flush_scheduled = False

    def _cleanup(self):
        # Make incoming logs flush immediately to ensure we communciate failures
        # to client on unexpected exit
        self.flush_after = None



from dagster_graphql.cli import execute_query

from dagster.core.errors import DagsterError

from .query import START_PIPELINE_EXECUTION_QUERY
from .util import HANDLED_EVENTS, dagster_event_from_dict


class DagsterGraphQLClientError(DagsterError):
    '''Indicates that some error has occurred when executing a GraphQL query against
    dagster-graphql'''


def execute_start_pipeline_execution_query(handle, variables):
    res = execute_query(
        handle,
        START_PIPELINE_EXECUTION_QUERY,
        variables,
        raise_on_error=True,
        use_sync_executor=True,
    )
    handle_start_pipeline_execution_errors(res)
    return handle_start_pipeline_execution_result(res)


def handle_start_pipeline_execution_errors(res):
    if res is None:
        raise DagsterGraphQLClientError('Unhandled error type. Raw response: {}'.format(res))

    if res.get('errors'):
        raise DagsterGraphQLClientError(
            'Internal error in GraphQL request. Response: {}'.format(res)
        )

    if not res.get('data', {}).get('startPipelineExecution', {}).get('__typename'):
        raise DagsterGraphQLClientError('Unexpected response type. Response: {}'.format(res))


def handle_start_pipeline_execution_result(res):
    res_data = res['data']['startPipelineExecution']

    res_type = res_data['__typename']

    if res_type == 'InvalidStepError':
        raise DagsterGraphQLClientError(
            'invalid step {step_key}'.format(step_key=res_data['invalidStepKey'])
        )

    if res_type == 'InvalidOutputError':
        raise DagsterGraphQLClientError(
            'invalid output {output} for step {step_key}'.format(
                output=res_data['invalidOutputName'], step_key=res_data['stepKey']
            )
        )

    if res_type == 'PipelineConfigValidationInvalid':
        errors = [err['message'] for err in res_data['errors']]
        raise DagsterGraphQLClientError(
            'Pipeline configuration invalid:\n{errors}'.format(errors='\n'.join(errors))
        )

    if res_type == 'PipelineNotFoundError':
        raise DagsterGraphQLClientError(
            'Pipeline "{pipeline_name}" not found: {message}:'.format(
                pipeline_name=res_data['pipelineName'], message=res_data['message']
            )
        )

    if res_type == 'PythonError':
        raise DagsterGraphQLClientError(
            'Subplan execution failed: {message}\n{stack}'.format(
                message=res_data['message'], stack=res_data['stack']
            )
        )

    if res_type == 'StartPipelineExecutionSuccess':
        pipeline_name = res_data['run']['pipeline']['name']

        skip_events = {
            'LogMessageEvent',
            'PipelineStartEvent',
            'PipelineSuccessEvent',
            'PipelineInitFailureEvent',
            'PipelineFailureEvent',
        }

        return [
            dagster_event_from_dict(e, pipeline_name)
            for e in res_data['run']['logs']['nodes']
            if e['__typename'] in HANDLED_EVENTS
        ]

    raise DagsterGraphQLClientError('Unexpected result type')
